
class LRMult(object):
    def __init__(self, lr_mult=1.):
        self.lr_mult = lr_mult

    def __call__(self, m):
        if getattr(m, 'weight', None) is not None:
            m.weight.lr_mult = self.lr_mult
        if getattr(m, 'bias', None) is not None:
            m.bias.lr_mult = self.lr_mult


from .qt import newAction, addActions, struct, newIcon
from .label import saveLabel, readLabel
from .config import parseConfigs, saveConfigs
from .network import model_path
from .colormap import ColorMask


import cv2
import random
import numpy as np

from albumentations.core.serialization import SERIALIZABLE_REGISTRY
from albumentations import ImageOnlyTransform, DualTransform
from albumentations.core.transforms_interface import to_tuple
from albumentations.augmentations import functional as F
from util.misc import get_bbox_from_mask, expand_bbox, clamp_bbox, get_labels_with_sizes


class UniformRandomResize(DualTransform):
    def __init__(self, scale_range=(0.9, 1.1), interpolation=cv2.INTER_LINEAR, always_apply=False, p=1):
        super().__init__(always_apply, p)
        self.scale_range = scale_range
        self.interpolation = interpolation

    def get_params_dependent_on_targets(self, params):
        scale = random.uniform(*self.scale_range)
        height = int(round(params['image'].shape[0] * scale))
        width = int(round(params['image'].shape[1] * scale))
        return {'new_height': height, 'new_width': width}

    def apply(self, img, new_height=0, new_width=0, interpolation=cv2.INTER_LINEAR, **params):
        return F.resize(img, height=new_height, width=new_width, interpolation=interpolation)

    def apply_to_keypoint(self, keypoint, new_height=0, new_width=0, **params):
        scale_x = new_width / params["cols"]
        scale_y = new_height / params["rows"]
        return F.keypoint_scale(keypoint, scale_x, scale_y)

    def get_transform_init_args_names(self):
        return "scale_range", "interpolation"

    @property
    def targets_as_params(self):
        return ["image"]


class ZoomIn(DualTransform):
    def __init__(
            self,
            height,
            width,
            bbox_jitter=0.1,
            expansion_ratio=1.4,
            min_crop_size=200,
            min_area=100,
            always_resize=False,
            always_apply=False,
            p=0.5,
    ):
        super(ZoomIn, self).__init__(always_apply, p)
        self.height = height
        self.width = width
        self.bbox_jitter = to_tuple(bbox_jitter)
        self.expansion_ratio = expansion_ratio
        self.min_crop_size = min_crop_size
        self.min_area = min_area
        self.always_resize = always_resize

    def apply(self, img, selected_object, bbox, **params):
        if selected_object is None:
            if self.always_resize:
                img = F.resize(img, height=self.height, width=self.width)
            return img

        rmin, rmax, cmin, cmax = bbox
        img = img[rmin:rmax + 1, cmin:cmax + 1]
        img = F.resize(img, height=self.height, width=self.width)

        return img

    def apply_to_mask(self, mask, selected_object, bbox, **params):
        if selected_object is None:
            if self.always_resize:
                mask = F.resize(mask, height=self.height, width=self.width,
                                interpolation=cv2.INTER_NEAREST)
            return mask

        rmin, rmax, cmin, cmax = bbox
        mask = mask[rmin:rmax + 1, cmin:cmax + 1]
        if isinstance(selected_object, tuple):
            layer_indx, mask_id = selected_object
            obj_mask = mask[:, :, layer_indx] == mask_id
            new_mask = np.zeros_like(mask)
            new_mask[:, :, layer_indx][obj_mask] = mask_id
        else:
            obj_mask = mask == selected_object
            new_mask = mask.copy()
            new_mask[np.logical_not(obj_mask)] = 0

        new_mask = F.resize(new_mask, height=self.height, width=self.width,
                            interpolation=cv2.INTER_NEAREST)
        return new_mask

    def get_params_dependent_on_targets(self, params):
        instances = params['mask']

        is_mask_layer = len(instances.shape) > 2
        candidates = []
        if is_mask_layer:
            for layer_indx in range(instances.shape[2]):
                labels, areas = get_labels_with_sizes(instances[:, :, layer_indx])
                candidates.extend([(layer_indx, obj_id)
                                   for obj_id, area in zip(labels, areas)
                                   if area > self.min_area])
        else:
            labels, areas = get_labels_with_sizes(instances)
            candidates = [obj_id for obj_id, area in zip(labels, areas)
                          if area > self.min_area]

        selected_object = None
        bbox = None
        if candidates:
            selected_object = random.choice(candidates)
            if is_mask_layer:
                layer_indx, mask_id = selected_object
                obj_mask = instances[:, :, layer_indx] == mask_id
            else:
                obj_mask = instances == selected_object

            bbox = get_bbox_from_mask(obj_mask)

            if isinstance(self.expansion_ratio, tuple):
                expansion_ratio = random.uniform(*self.expansion_ratio)
            else:
                expansion_ratio = self.expansion_ratio

            bbox = expand_bbox(bbox, expansion_ratio, self.min_crop_size)
            bbox = self._jitter_bbox(bbox)
            bbox = clamp_bbox(bbox, 0, obj_mask.shape[0] - 1, 0, obj_mask.shape[1] - 1)

        return {
            'selected_object': selected_object,
            'bbox': bbox
        }

    def _jitter_bbox(self, bbox):
        rmin, rmax, cmin, cmax = bbox
        height = rmax - rmin + 1
        width = cmax - cmin + 1
        rmin = int(rmin + random.uniform(*self.bbox_jitter) * height)
        rmax = int(rmax + random.uniform(*self.bbox_jitter) * height)
        cmin = int(cmin + random.uniform(*self.bbox_jitter) * width)
        cmax = int(cmax + random.uniform(*self.bbox_jitter) * width)

        return rmin, rmax, cmin, cmax

    def apply_to_bbox(self, bbox, **params):
        raise NotImplementedError

    def apply_to_keypoint(self, keypoint, **params):
        raise NotImplementedError

    @property
    def targets_as_params(self):
        return ["mask"]

    def get_transform_init_args_names(self):
        return ("height", "width", "bbox_jitter",
                "expansion_ratio", "min_crop_size", "min_area", "always_resize")


def remove_image_only_transforms(sdict):
    if not 'transforms' in sdict:
        return sdict

    keep_transforms = []
    for tdict in sdict['transforms']:
        cls = SERIALIZABLE_REGISTRY[tdict['__class_fullname__']]
        if 'transforms' in tdict:
            keep_transforms.append(remove_image_only_transforms(tdict))
        elif not issubclass(cls, ImageOnlyTransform):
            keep_transforms.append(tdict)
    sdict['transforms'] = keep_transforms

    return sdict


import os
import os.path as osp


def toint(seq):
    for idx in range(len(seq)):
        try:
            seq[idx] = int(seq[idx])
        except ValueError:
            pass
    return seq


def saveLabel(labelList, path):
    # labelList = [[1, "人", [0, 0, 0]], [2, "车", [128, 128, 128]]]
    print("save label", labelList, path)
    print(osp.exists(osp.dirname(path)), osp.dirname(path))
    if not path or len(path) == 0 or not osp.exists(osp.dirname(path)):
        print("save label error")
        return
    with open(path, "w", encoding="utf-8") as f:
        for l in labelList:
            for idx in range(2):
                print(l[idx], end=" ", file=f)
            for idx in range(3):
                print(l[2][idx], end=" ", file=f)
            print(file=f)


# saveLabel("label.txt")


def readLabel(path):
    if not path or len(path) == 0 or not osp.exists(path):
        return []

    with open(path, "r", encoding="utf-8") as f:
        labels = f.readlines()
    labelList = []
    for lab in labels:
        lab = lab.replace("\n", "").strip(" ").split(" ")
        if len(lab) != 2 and len(lab) != 5:
            print("标签不合法")
            continue
        label = toint(lab[:2])
        label.append(toint(lab[2:]))
        labelList.append(label)
    print(labelList)
    return labelList


# readLabel("label.txt")



import time

import paddle
# from tkinter import messagebox

import numpy as np
import paddleseg.transforms as T

from inference import clicker
from inference.predictor import get_predictor
from util.vis import draw_with_blend_and_clicks


class InteractiveController:
    def __init__(self, net, predictor_params, update_image_callback, prob_thresh=0.5):
        self.net = net
        self.prob_thresh = prob_thresh
        self.clicker = clicker.Clicker()
        self.states = []
        self.probs_history = []
        self.curr_label_number = 0
        self._result_mask = None
        self.label_list = None  # 存标签编号和颜色的对照
        self._init_mask = None

        self.image = None
        self.predictor = None
        self.update_image_callback = update_image_callback
        self.predictor_params = predictor_params
        self.reset_predictor()

    def set_image(self, image):
        """设置当前标注的图片

        Parameters
        ----------
        image :
            Description of parameter `image`.
        """
        # TODO: 这里normalize需要按照模型改
        # input_transform = T.Compose(
        #     [T.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])],
        #     to_rgb=False,
        # )
        self.image = image
        # self.image_nd = input_transform(image)[0]
        self._result_mask = np.zeros(image.shape[:2], dtype=np.uint8)
        # self.curr_label_number = 0
        self.reset_last_object(update_image=False)
        self.update_image_callback(reset_canvas=True)

    def set_mask(self, mask):
        # if self.image.shape[:2] != mask.shape[:2]:
        #     messagebox.showwarning(
        #         "Warning",
        #         "A segmentation mask must have the same sizes as the current image!",
        #     )
        #     return

        if len(self.probs_history) > 0:
            self.reset_last_object()

        self._init_mask = mask.astype(np.float32)
        self.probs_history.append((np.zeros_like(self._init_mask), self._init_mask))
        self._init_mask = paddle.to_tensor(self._init_mask).unsqueeze(0).unsqueeze(0)
        self.clicker.click_indx_offset = 1

    def add_click(self, x, y, is_positive):
        """添加一个点
        跑推理，保存历史用于undo
        Parameters
        ----------
        x : type
            Description of parameter `x`.
        y : type
            Description of parameter `y`.
        is_positive : bool
            是否是正点
        Returns
            -------
            bool
                点击是否成功添加
        """
        s = self.image.shape
        if x < 0 or y < 0 or x >= s[1] or y >= s[0]:
            print("点击越界")
            return False
        self.states.append(
            {
                "clicker": self.clicker.get_state(),
                "predictor": self.predictor.get_states(),
            }
        )

        click = clicker.Click(is_positive=is_positive, coords=(y, x))
        self.clicker.add_click(click)
        start = time.time()
        print(self.predictor)
        pred = self.predictor.get_prediction(self.clicker, prev_mask=self._init_mask)
        if self._init_mask is not None and len(self.clicker) == 1:
            pred = self.predictor.get_prediction(
                self.clicker, prev_mask=self._init_mask
            )
        end = time.time()
        print("cost time", end - start)

        if self.probs_history:
            self.probs_history.append((self.probs_history[-1][0], pred))
        else:
            self.probs_history.append((np.zeros_like(pred), pred))

        self.update_image_callback()

    def set_label(self, label):
        # if label is None:
        #     return
        # self.probs_history.append((np.zeros_like(label), label))
        # print("len", len(self.probs_history))
        # self.update_image_callback()
        pass

    def undo_click(self):
        """undo一步点击"""
        if not self.states:  # 如果还没点
            return

        prev_state = self.states.pop()
        self.clicker.set_state(prev_state["clicker"])
        self.predictor.set_states(prev_state["predictor"])
        self.probs_history.pop()
        if not self.probs_history:
            self.reset_init_mask()
        self.update_image_callback()

    def partially_finish_object(self):
        """部分完成
        保存一个mask的状态，这个状态里不存点，看起来比较
        """

        object_prob = self.current_object_prob
        if object_prob is None:
            return

        self.probs_history.append((object_prob, np.zeros_like(object_prob)))
        self.states.append(self.states[-1])

        self.clicker.reset_clicks()
        self.reset_predictor()
        self.reset_init_mask()
        self.update_image_callback()

    def finish_object(self):
        """结束当前物体标注，准备标下一个"""
        object_prob = self.current_object_prob
        if object_prob is None:
            return

        # self.curr_label_number += 1  # TODO: 当前是按照第几个目标给结果中的数，改成根据目标编号
        object_mask = object_prob > self.prob_thresh
        print('curr_label_number:', self.curr_label_number)
        self._result_mask[object_mask] = self.curr_label_number
        self.reset_last_object()

    def change_label_num(self, number):
        """修改当前标签的编号
        如果当前有标注到一半的目标，改mask。
        如果没有，下一个目标是这个数
        Parameters
        ----------
        number : int
            换成目标的编号
        """
        assert isinstance(number, int), "标签编号应为整数"
        self.curr_label_number = number
        if self.is_incomplete_mask:
            pass
            # TODO: 改当前mask的编号

    def reset_last_object(self, update_image=True):
        """重置控制器状态
        Parameters
        ----------
        update_image : bool
            Description of parameter `update_image`.
        Returns
        -------
        type
            Description of returned object.

        """
        self.states = []
        self.probs_history = []
        self.clicker.reset_clicks()
        self.reset_predictor()
        self.reset_init_mask()
        if update_image:
            self.update_image_callback()

    def reset_predictor(self, net=None, predictor_params=None):
        """重置推理器，可以换权重
        Parameters
        ----------
        predictor_params : 网络权重
            新的网络权重
        """
        # print("palette", self.palette)
        if net is not None:
            self.net = net
        if predictor_params is not None:
            self.predictor_params = predictor_params
        self.predictor = get_predictor(self.net, **self.predictor_params)
        if self.image is not None:
            self.predictor.set_input_image(self.image)

    def reset_init_mask(self):
        self._init_mask = None
        self.clicker.click_indx_offset = 0

    @property
    def current_object_prob(self):
        if self.probs_history:
            current_prob_total, current_prob_additive = self.probs_history[-1]
            return np.maximum(current_prob_total, current_prob_additive)
        else:
            return None

    @property
    def is_incomplete_mask(self):
        return len(self.probs_history) > 0

    @property
    def result_mask(self):
        result_mask = self._result_mask.copy()
        if self.probs_history:
            result_mask[self.current_object_prob > self.prob_thresh] = (
                self.object_count + 1
            )
        return result_mask

    def get_visualization(self, alpha_blend, click_radius):
        if self.image is None:
            return None

        # 1. 画当前没标完的mask
        results_mask_for_vis = self.result_mask
        if self.probs_history:
            results_mask_for_vis[
                self.current_object_prob > self.prob_thresh
            ] = self.curr_label_number

        vis = draw_with_blend_and_clicks(
            self.image,
            mask=results_mask_for_vis,
            alpha=alpha_blend,
            clicks_list=self.clicker.clicks_list,
            radius=click_radius,
            palette=self.palette,
        )

        # 2. 在图片和当前mask的基础上画之前标完的mask
        if self.probs_history:
            total_mask = self.probs_history[-1][0] > self.prob_thresh
            results_mask_for_vis[np.logical_not(total_mask)] = 0
            vis = draw_with_blend_and_clicks(
                vis,
                mask=results_mask_for_vis,
                alpha=alpha_blend,
                palette=self.palette,
            )

        return vis

    @property
    def palette(self):
        if self.label_list:
            colors = [l[2] for l in self.label_list]
            colors.insert(0, [0, 0, 0])
        else:
            colors = [[0, 0, 0]]
        print(colors)
        return colors

    @property
    def current_object_prob(self):
        """获取当前推理标签"""
        if self.probs_history:
            current_prob_total, current_prob_additive = self.probs_history[-1]
            return np.maximum(current_prob_total, current_prob_additive)
        else:
            return None

    @property
    def is_incomplete_mask(self):
        """
        Returns
        -------
        bool
            当前的物体是不是还没标完
        """
        return len(self.probs_history) > 0

    @property
    def result_mask(self):
        return self._result_mask.copy()

    @property
    def img_size(self):
        print(self.image.shape)
        return self.image.shape[1::-1]



import paddle
import paddle.nn.functional as F
import numpy as np

from model.metrics import _compute_iou
from .brs_losses import BRSMaskLoss


class BaseOptimizer:
    def __init__(self, optimizer_params,
                 prob_thresh=0.49,
                 reg_weight=1e-3,
                 min_iou_diff=0.01,
                 brs_loss=BRSMaskLoss(),
                 with_flip=False,
                 flip_average=False,
                 **kwargs):
        self.brs_loss = brs_loss
        self.optimizer_params = optimizer_params
        self.prob_thresh = prob_thresh
        self.reg_weight = reg_weight
        self.min_iou_diff = min_iou_diff
        self.with_flip = with_flip
        self.flip_average = flip_average

        self.best_prediction = None
        self._get_prediction_logits = None
        self._opt_shape = None
        self._best_loss = None
        self._click_masks = None
        self._last_mask = None

    def init_click(self, get_prediction_logits, pos_mask, neg_mask, shape=None):
        self.best_prediction = None
        self._get_prediction_logits = get_prediction_logits
        self._click_masks = (pos_mask, neg_mask)
        self._opt_shape = shape
        self._last_mask = None

    def __call__(self, x):
        opt_params = paddle.to_tensor(x).astype('float64')
        opt_params.stop_gradient = False   
#         with enable_grad():
#             scale, bias, reg_loss = self.unpack_opt_params(opt_params)
#             result_before_sigmoid = self._get_prediction_logits(scale, bias)
#             result = F.sigmoid(result_before_sigmoid)
#             pos_mask, neg_mask = self._click_masks
#             if self.with_flip and self.flip_average:
#                 result, result_flipped = paddle.chunk(result, 2, axis=0)
#                 result = 0.5 * (result + paddle.flip(result_flipped, axis=[3]))
#                 pos_mask, neg_mask = pos_mask[:result.shape[0]], neg_mask[:result.shape[0]]

#             loss, f_max_pos, f_max_neg = self.brs_loss(result, pos_mask, neg_mask)
#             loss = loss + reg_loss

        opt_vars, reg_loss = self.unpack_opt_params(opt_params)
        result_before_sigmoid = self._get_prediction_logits(*opt_vars)
        result = F.sigmoid(result_before_sigmoid)

        pos_mask, neg_mask = self._click_masks
        if self.with_flip and self.flip_average:
            result, result_flipped = paddle.chunk(result, 2, axis=0)
            result = 0.5 * (result + paddle.flip(result_flipped, axis=[3]))
            pos_mask, neg_mask = pos_mask[:result.shape[0]], neg_mask[:result.shape[0]]

        loss, f_max_pos, f_max_neg = self.brs_loss(result, pos_mask, neg_mask)
        loss = loss + reg_loss

        f_val = loss.detach().numpy()
        if self.best_prediction is None or f_val < self._best_loss:
            self.best_prediction = result_before_sigmoid.detach()
            self._best_loss = f_val
        if f_max_pos < (1 - self.prob_thresh) and f_max_neg < self.prob_thresh:
            return [f_val, np.zeros_like(x)]

        current_mask = result > self.prob_thresh
        if self._last_mask is not None and self.min_iou_diff > 0:
            diff_iou = _compute_iou(current_mask, self._last_mask)
            if len(diff_iou) > 0 and diff_iou.mean() > 1 - self.min_iou_diff:
                return [f_val, np.zeros_like(x)]
        self._last_mask = current_mask
        
        loss.backward()
        f_grad = opt_params.gradient().ravel().astype(np.float64)

        return [f_val, f_grad]

    def unpack_opt_params(self, opt_params):
        raise NotImplementedError


class InputOptimizer(BaseOptimizer):
    def unpack_opt_params(self, opt_params):
        opt_params = opt_params.reshape(self._opt_shape)
        if self.with_flip:
            opt_params_flipped = paddle.flip(opt_params, axis=[3])
            opt_params = paddle.concat([opt_params, opt_params_flipped], axis=0)
        reg_loss = self.reg_weight * paddle.sum(opt_params ** 2)

        return (opt_params,), reg_loss


class ScaleBiasOptimizer(BaseOptimizer):
    def __init__(self, *args, scale_act=None, reg_bias_weight=10.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.scale_act = scale_act
        self.reg_bias_weight = reg_bias_weight

    def unpack_opt_params(self, opt_params):
        scale, bias = paddle.chunk(opt_params, 2, axis=0)
        reg_loss = self.reg_weight * (paddle.sum(scale ** 2) + self.reg_bias_weight * paddle.sum(bias ** 2))

        if self.scale_act == 'tanh':
            scale = paddle.tanh(scale)
        elif self.scale_act == 'sin':
            scale = paddle.sin(scale)

        return (1 + scale, bias), reg_loss



from pathlib import Path
import pickle
import random
import numpy as np
import json
import cv2
from copy import deepcopy
from data.base import ISDataset
from data.sample import DSample


class CocoLvisDataset(ISDataset):
    def __init__(self, dataset_path, split='train', stuff_prob=0.0,
                 allow_list_name=None, anno_file='hannotation.pickle', **kwargs):
        super(CocoLvisDataset, self).__init__(**kwargs)
        dataset_path = Path(dataset_path)
        self._split_path = dataset_path / split
        self.split = split
        self._images_path = self._split_path / 'images'
        self._masks_path = self._split_path / 'masks'
        self.stuff_prob = stuff_prob

        with open(self._split_path / anno_file, 'rb') as f:
            self.dataset_samples = sorted(pickle.load(f).items())

        if allow_list_name is not None:
            allow_list_path = self._split_path / allow_list_name
            with open(allow_list_path, 'r') as f:
                allow_images_ids = json.load(f)
            allow_images_ids = set(allow_images_ids)

            self.dataset_samples = [sample for sample in self.dataset_samples
                                    if sample[0] in allow_images_ids]

    def get_sample(self, index) -> DSample:
        # 将mask都读取出来，然后取出label之后单独处理，使得每个instance得到一个单独都mask，物体部分为0，生成一个binary mask。
        image_id, sample = self.dataset_samples[index]
        image_path = self._images_path / f'{image_id}.jpg'

        image = cv2.imread(str(image_path))
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        packed_masks_path = self._masks_path / f'{image_id}.pickle'
        with open(packed_masks_path, 'rb') as f:
            encoded_layers, objs_mapping = pickle.load(f)
        layers = [cv2.imdecode(x, cv2.IMREAD_UNCHANGED) for x in encoded_layers]
        layers = np.stack(layers, axis=2)

        instances_info = deepcopy(sample['hierarchy'])
        for inst_id, inst_info in list(instances_info.items()):
            if inst_info is None:
                inst_info = {'children': [], 'parent': None, 'node_level': 0}
                instances_info[inst_id] = inst_info
            inst_info['mapping'] = objs_mapping[inst_id]

        if self.stuff_prob > 0 and random.random() < self.stuff_prob:
            for inst_id in range(sample['num_instance_masks'], len(objs_mapping)):
                instances_info[inst_id] = {
                    'mapping': objs_mapping[inst_id],
                    'parent': None,
                    'children': []
                }
        else:
            #mask中有4个layer，每个layer标着在各自的layer上都有什么mask—id
            for inst_id in range(sample['num_instance_masks'], len(objs_mapping)):
                layer_indx, mask_id = objs_mapping[inst_id]
                layers[:, :, layer_indx][layers[:, :, layer_indx] == mask_id] = 0

        return DSample(image, layers, objects=instances_info)


import yaml
import os.path as osp


def parseConfigs(path):
    if not path or not osp.exists(path):
        return
    with open(path, "r", encoding="utf-8") as f:
        return yaml.load(f.read(), Loader=yaml.FullLoader)


def saveConfigs(path, config):
    if not osp.exists(osp.basename(path)):
        print("文件夹不存在")
        return
    with open(path, "w", encoding="utf-8") as f:
        yaml.dump(config, f)


class cfgData(object):
    def __init__(self, yaml_file):
        with open(yaml_file, "r", encoding="utf-8") as f:
            fig_data = f.read()
            self.dicts = yaml.load(fig_data)

    def get(self, key):
        if key in self.dicts.keys():
            return self.dicts[key]
        else:
            raise ValueError("Not find this keyword.")


if __name__ == "__main__":
    cfg = cfgData("EISeg/train/train_config.yaml")
    print(cfg.get("use_vdl"))



import random
import pickle

import cv2
import numpy as np
import paddle

import paddleseg.transforms as T
from .points_sampler import MultiPointSampler


def get_unique_labels(x, exclude_zero=False):
    obj_sizes = np.bincount(x.flatten())
    labels = np.nonzero(obj_sizes)[0].tolist()

    if exclude_zero:
        labels = [x for x in labels if x != 0]
    return labels


class ISDataset(paddle.io.Dataset):
    def __init__(self,
                 augmentator=None,
                 points_sampler=MultiPointSampler(max_num_points=12),
                 min_object_area=0,
                 min_ignore_object_area=10,
                 keep_background_prob=0.0,
                 with_image_info=False,
                 samples_scores_path=None,
                 samples_scores_gamma=1.0,
                 epoch_len=-1):
        super(ISDataset, self).__init__()
        self.epoch_len = epoch_len
        self.augmentator = augmentator
        self.min_object_area = min_object_area
        self.keep_background_prob = keep_background_prob
        self.points_sampler = points_sampler
        self.with_image_info = with_image_info
        self.samples_precomputed_scores = self._load_samples_scores(samples_scores_path, samples_scores_gamma)
        self.dataset_samples = None
        
    def to_tensor(self, x):
        if isinstance(x, np.ndarray):
            if x.ndim == 2:
                x = x[:,:,None]
        #img = paddle.to_tensor(x.transpose([2,0,1])).astype('float32') / 255
        img = x.transpose([2,0,1]).astype(np.float32) / 255
        return img

    def __getitem__(self, index):
        
#         if self.samples_precomputed_scores is not None:
#             index = np.random.choice(self.samples_precomputed_scores['indices'],
#                                      p=self.samples_precomputed_scores['probs'])
#         else:
#             if self.epoch_len > 0:
#                 index = random.randrange(0, len(self.dataset_samples))
        sample = self.get_sample(index)
        sample = self.augment_sample(sample)
        sample.remove_small_objects(self.min_object_area)
        self.points_sampler.sample_object(sample)
        points = np.array(self.points_sampler.sample_points()).astype(np.float32)
        mask = self.points_sampler.selected_mask
        image = self.to_tensor(sample.image)
        ids = sample.sample_id

        return image, points, mask

    def augment_sample(self, sample):
        if self.augmentator is None:
            return sample

        valid_augmentation = False
        while not valid_augmentation:
            sample.augment(self.augmentator)
            keep_sample = (self.keep_background_prob < 0.0 or
                           random.random() < self.keep_background_prob)
            valid_augmentation = len(sample) > 0 or keep_sample

        return sample

    def get_sample(self, index):
        raise NotImplementedError

    def __len__(self):
        if self.epoch_len > 0:
            return self.epoch_len
        else:
            return self.get_samples_number()

    def get_samples_number(self):
        return len(self.dataset_samples)

    @staticmethod
    def _load_samples_scores(samples_scores_path, samples_scores_gamma):
        if samples_scores_path is None:
            return None

        with open(samples_scores_path, 'rb') as f:
            images_scores = pickle.load(f)

        probs = np.array([(1.0 - x[2]) ** samples_scores_gamma for x in images_scores])
        probs /= probs.sum()
        samples_scores = {
            'indices': [x[0] for x in images_scores],
            'probs': probs
        }
        print(f'Loaded {len(probs)} weights with gamma={samples_scores_gamma}')
        return samples_scores



from pathlib import Path

import os
import cv2
import numpy as np

from data.base import ISDataset
from data.sample import DSample


class HumanDataset(ISDataset):
    def __init__(self, dataset_path,
                 split = 'train',
                 **kwargs):
        super(HumanDataset, self).__init__(**kwargs)
        
        self.mode = split.lower()
        self.path = dataset_path

        if self.mode == 'train':
            file_path = os.path.join(self.path, 'train_mini.txt')
        else:
            file_path = os.path.join(self.path, 'val_mini.txt')

        self.dataset_samples = []
        with open(file_path, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line != '':
                    self.dataset_samples.append(line)

    def get_sample(self, index):
        items = self.dataset_samples[index].split(' ')
        if 'person_detection__ds' in items[0]:
            image_path, image_name = items[0].rsplit('/', 1)
            items[0] = image_path.rsplit('/', 1)[0] + '/' + image_name
            mask_path, mask_name = items[1].rsplit('/', 1)
            items[1] = mask_path.rsplit('/', 1)[0] + '/' + mask_name
        

        image_path = os.path.join(self.path, items[0])
        mask_path = os.path.join(self.path, items[1])

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        instances_mask = np.max(cv2.imread(mask_path).astype(np.int32), axis=2)
        instances_mask[instances_mask > 0] = 1
       
        return  DSample(image, instances_mask, objects_ids=[1], sample_id=index)




import paddle.nn as nn

from util.serialization import serialize
from .is_model import ISModel
from .modeling.deeplab_v3 import DeepLabV3Plus
from .modeling.basic_blocks import SepConvHead
from model.modifiers import LRMult

class DeeplabModel(ISModel):
    @serialize
    def __init__(self, backbone='resnet50', deeplab_ch=256, aspp_dropout=0.5,
                 backbone_norm_layer=None, backbone_lr_mult=0.1, norm_layer=nn.BatchNorm2D, **kwargs):
        super().__init__(norm_layer=norm_layer, **kwargs)

        self.feature_extractor = DeepLabV3Plus(backbone=backbone, ch=deeplab_ch, project_dropout=aspp_dropout,
                                               norm_layer=norm_layer, backbone_norm_layer=backbone_norm_layer)
        self.feature_extractor.backbone.apply(LRMult(backbone_lr_mult))
        self.head = SepConvHead(1, in_channels=deeplab_ch, mid_channels=deeplab_ch // 2,
                                num_layers=2, norm_layer=norm_layer)

    def backbone_forward(self, image, coord_features=None):
        backbone_features = self.feature_extractor(image, coord_features)

        return {'instances': self.head(backbone_features[0])}


import paddle.nn as nn

from util.serialization import serialize
from .is_model import ISModel
from .modeling.hrnet_ocr import HighResolutionNet
from model.modifiers import LRMult


class HRNetModel(ISModel):
    def __init__(self, width=48, ocr_width=256, small=False, backbone_lr_mult=0.1,
                 norm_layer=nn.BatchNorm2D, **kwargs):
        super().__init__(norm_layer=norm_layer, **kwargs)

        self.feature_extractor = HighResolutionNet(width=width, ocr_width=ocr_width, small=small,
                                                   num_classes=1, norm_layer=norm_layer)
        self.feature_extractor.apply(LRMult(backbone_lr_mult))
        if ocr_width > 0:
            self.feature_extractor.ocr_distri_head.apply(LRMult(1.0))
            self.feature_extractor.ocr_gather_head.apply(LRMult(1.0))
            self.feature_extractor.conv3x3_ocr.apply(LRMult(1.0))

    def backbone_forward(self, image, coord_features=None):
        net_outputs = self.feature_extractor(image, coord_features)

        return {'instances': net_outputs[0], 'instances_aux': net_outputs[1]}



from datetime import timedelta
from pathlib import Path
import paddle

import numpy as np

from data.datasets import GrabCutDataset, BerkeleyDataset, DavisDataset, SBDEvaluationDataset, PascalVocDataset, HumanDataset
from util.serialization import load_model


def get_time_metrics(all_ious, elapsed_time):
    n_images = len(all_ious)
    n_clicks = sum(map(len, all_ious))

    mean_spc = elapsed_time / n_clicks
    mean_spi = elapsed_time / n_images

    return mean_spc, mean_spi


def load_is_model(checkpoint, **kwargs):
    if isinstance(checkpoint, (str, Path)):
        state_dict = paddle.load(checkpoint)
    else:
        state_dict = checkpoint

    if isinstance(state_dict, list):
        model = load_single_is_model(state_dict[0],  **kwargs)
        models = [load_single_is_model(x, **kwargs) for x in state_dict]

        return model, models
    else:
        return load_single_is_model(state_dict, **kwargs)


def load_single_is_model(state_dict, **kwargs):
    model = load_model(state_dict['config'], **kwargs)
    model.load_state_dict(state_dict['state_dict'], strict=False)

    for param in model.parameters():
        param.requires_grad = False
    model.eval()

    return model


def get_dataset(dataset_name, cfg):
    if dataset_name == 'GrabCut':
        dataset = GrabCutDataset(cfg.GRABCUT_PATH)
    elif dataset_name == 'Berkeley':
        dataset = BerkeleyDataset(cfg.BERKELEY_PATH)
    elif dataset_name == 'DAVIS':
        dataset = DavisDataset(cfg.DAVIS_PATH)
    elif dataset_name == 'SBD':
        dataset = SBDEvaluationDataset(cfg.SBD_PATH)
    elif dataset_name == 'SBD_Train':
        dataset = SBDEvaluationDataset(cfg.SBD_PATH, split='train')
    elif dataset_name == 'PascalVOC':
        dataset = PascalVocDataset(cfg.PASCALVOC_PATH, split='val')
    elif dataset_name == 'COCO_MVal':
        dataset = DavisDataset(cfg.COCO_MVAL_PATH)
    elif dataset_name == 'Human':
        dataset = HumanDataset(cfg.HUMAN_PATH, split='test')
    else:
        dataset = None

    return dataset


def get_iou(gt_mask, pred_mask, ignore_label=-1):
    ignore_gt_mask_inv = gt_mask != ignore_label
    obj_gt_mask = gt_mask == 1

    intersection = np.logical_and(np.logical_and(pred_mask, obj_gt_mask), ignore_gt_mask_inv).sum()
    union = np.logical_and(np.logical_or(pred_mask, obj_gt_mask), ignore_gt_mask_inv).sum()

    return intersection / union


def compute_noc_metric(all_ious, iou_thrs, max_clicks=20):
    def _get_noc(iou_arr, iou_thr):
        vals = iou_arr >= iou_thr
        return np.argmax(vals) + 1 if np.any(vals) else max_clicks

    noc_list = []
    over_max_list = []
    for iou_thr in iou_thrs:
        scores_arr = np.array([_get_noc(iou_arr, iou_thr)
                               for iou_arr in all_ious], dtype=np.int)

        score = scores_arr.mean()
        over_max = (scores_arr == max_clicks).sum()

        noc_list.append(score)
        over_max_list.append(over_max)

    return noc_list, over_max_list


def find_checkpoint(weights_folder, checkpoint_name):
    weights_folder = Path(weights_folder)
    if ':' in checkpoint_name:
        model_name, checkpoint_name = checkpoint_name.split(':')
        models_candidates = [x for x in weights_folder.glob(f'{model_name}*') if x.is_dir()]
        assert len(models_candidates) == 1
        model_folder = models_candidates[0]
    else:
        model_folder = weights_folder

    if checkpoint_name.endswith('.pth'):
        if Path(checkpoint_name).exists():
            checkpoint_path = checkpoint_name
        else:
            checkpoint_path = weights_folder / checkpoint_name
    else:
        model_checkpoints = list(model_folder.rglob(f'{checkpoint_name}*.pdparams'))
        assert len(model_checkpoints) == 1
        checkpoint_path = model_checkpoints[0]

    return str(checkpoint_path)


def get_results_table(noc_list, over_max_list, brs_type, dataset_name, mean_spc, elapsed_time,
                      n_clicks=20, model_name=None):
    table_header = (f'|{"BRS Type":^13}|{"Dataset":^11}|'
                    f'{"NoC@80%":^9}|{"NoC@85%":^9}|{"NoC@90%":^9}|'
                    f'{">="+str(n_clicks)+"@85%":^9}|{">="+str(n_clicks)+"@90%":^9}|'
                    f'{"SPC,s":^7}|{"Time":^9}|')
    row_width = len(table_header)

    header = f'Eval results for model: {model_name}\n' if model_name is not None else ''
    header += '-' * row_width + '\n'
    header += table_header + '\n' + '-' * row_width

    eval_time = str(timedelta(seconds=int(elapsed_time)))
    table_row = f'|{brs_type:^13}|{dataset_name:^11}|'
    table_row += f'{noc_list[0]:^9.2f}|'
    table_row += f'{noc_list[1]:^9.2f}|' if len(noc_list) > 1 else f'{"?":^9}|'
    table_row += f'{noc_list[2]:^9.2f}|' if len(noc_list) > 2 else f'{"?":^9}|'
    table_row += f'{over_max_list[1]:^9}|' if len(noc_list) > 1 else f'{"?":^9}|'
    table_row += f'{over_max_list[2]:^9}|' if len(noc_list) > 2 else f'{"?":^9}|'
    table_row += f'{mean_spc:^7.3f}|{eval_time:^9}|'

    return header, table_row


import pickle as pkl
from pathlib import Path
from random import choice

import cv2
import numpy as np

from data.base import ISDataset
from data.sample import DSample


class PascalVocDataset(ISDataset):
    def __init__(self, dataset_path, split='train', **kwargs):
        super().__init__(**kwargs)
        assert split in {'train', 'val', 'trainval'}

        self.dataset_path = Path(dataset_path)
        self._images_path = self.dataset_path / "JPEGImages"
        self._insts_path = self.dataset_path / "SegmentationObject"
        self.dataset_split = split
        with open(self.dataset_path / f'ImageSets/Segmentation/{split}.txt', 'r') as f:
            
            self.dataset_samples = [name.strip() for name in f.readlines()]

    def get_sample(self, index) -> DSample:
        sample_id = self.dataset_samples[index]
        image_path = str(self._images_path / f'{sample_id}.jpg')
        mask_path = str(self._insts_path / f'{sample_id}.png')

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        instances_mask = cv2.imread(mask_path)
        instances_mask = cv2.cvtColor(instances_mask, cv2.COLOR_BGR2GRAY).astype(np.int32)
        if self.dataset_split == 'val':
            objects_ids = np.unique(instances_mask)
            objects_ids = [x for x in objects_ids if x != 0 and x != 220]
            instance_id = choice(objects_ids)
            mask = np.zeros_like(instances_mask)
            mask[instances_mask == 220] = 220  # ignored area
            mask[instances_mask == instance_id] = 1
            objects_ids = [1]
            instances_mask = mask
        else:
            objects_ids = np.unique(instances_mask)
            objects_ids = [x for x in objects_ids if x != 0 and x != 220]

        return DSample(image, instances_mask, objects_ids=objects_ids, ignore_ids=[220], sample_id=index)



import paddle
import paddle.nn as nn
import numpy as np
from paddle.fluid.initializer import  Initializer

class SInitializer(Initializer):
    def __init__(self, local_init=True, gamma=None):
        self.local_init = local_init
        self.gamma = gamma

    def __call__(self, m):


        if isinstance(m, (nn.BatchNorm1D, nn.BatchNorm2D, nn.BatchNorm3D,
                          nn.InstanceNorm1D, nn.InstanceNorm2D, nn.InstanceNorm3D,
                          nn.GroupNorm, nn.SyncBatchNorm)) or 'BatchNorm' in m.__class__.__name__:
            if m.weight is not None:
                self._init_gamma(m.weight)
            if m.bias is not None:
                self._init_beta(m.bias)
            else:
                if getattr(m, 'weight', None) is not None:
                    self._init_weight(m.weight)
                if getattr(m, 'bias', None) is not None:
                    self._init_bias(m.bias)

    def _init_weight(self, param):
        initializer = nn.initializer.Uniform(-0.07, 0.07)
        initializer(param, param.block)

    def _init_bias(self, param):
        initializer = nn.initializer.Constant(0)
        initializer(param, param.block)

    def _init_gamma(self, param):
        if self.gamma is None:
            initializer = nn.initializer.Constant(0)
            initializer(param, param.block)
        else:
            initializer = nn.initializer.Normal(1, self.gamma)
            initializer(param, param.block)

    def _init_beta(self, param):
        initializer = nn.initializer.Constant(0)
        initializer(param, param.block)

        
class XavierGluon(SInitializer):
    def __init__(self, rnd_type='uniform', factor_type='avg', magnitude=3, **kwargs):
        super().__init__(**kwargs)

        self.rnd_type = rnd_type
        self.factor_type = factor_type
        self.magnitude = float(magnitude)

    def _init_weight(self, arr):
        fan_in, fan_out =self._compute_fans(arr)

        if self.factor_type == 'avg':
            factor = (fan_in + fan_out) / 2.0
        elif self.factor_type == 'in':
            factor = fan_in
        elif self.factor_type == 'out':
            factor = fan_out
        else:
            raise ValueError('Incorrect factor type')
        scale = np.sqrt(self.magnitude / factor)

        if self.rnd_type == 'uniform':
            initializer = nn.initializer.Uniform(-scale, scale)
            initializer(arr, arr.block)
        elif self.rnd_type == 'gaussian':
            initializer = nn.initializer.Normal(0, scale)
            initializer(arr, arr.block)
        else:
            raise ValueError('Unknown random type')



from contextlib import ExitStack

import paddle
import paddle.nn as nn
import paddle.nn.functional as F

from .basic_blocks import SeparableConv2d
from paddleseg.models.backbones import *
from .resnet import ResNetBackbone

class DeepLabV3Plus(nn.Layer):
    def __init__(self, backbone='resnet50', norm_layer=nn.BatchNorm2D,
                 backbone_norm_layer=None,
                 ch=256,
                 project_dropout=0.5,
                 inference_mode=False,
                 **kwargs):
        super(DeepLabV3Plus, self).__init__()
        if backbone_norm_layer is None:
            backbone_norm_layer = norm_layer

        self.backbone_name = backbone
        self.norm_layer = norm_layer
        self.backbone_norm_layer = backbone_norm_layer
        self.inference_mode = False
        self.ch = ch
        self.aspp_in_channels = 2048
        self.skip_project_in_channels = 256  # layer 1 out_channels

        self._kwargs = kwargs
        if backbone == 'resnet34':
            self.aspp_in_channels = 512
            self.skip_project_in_channels = 64

        self.backbone = ResNetBackbone(backbone=self.backbone_name, pretrained_base=False,
                                       norm_layer=self.backbone_norm_layer, **kwargs)
        
        
        self.head = _DeepLabHead(in_channels=ch + 32, mid_channels=ch, out_channels=ch,
                                 norm_layer=self.norm_layer)
        self.skip_project = _SkipProject(self.skip_project_in_channels, 32, norm_layer=self.norm_layer)
        self.aspp = _ASPP(in_channels=self.aspp_in_channels,
                          atrous_rates=[12, 24, 36],
                          out_channels=ch,
                          project_dropout=project_dropout,
                          norm_layer=self.norm_layer)

        if inference_mode:
            self.set_prediction_mode()

    def load_pretrained_weights(self):
        pretrained = ResNetBackbone(backbone=self.backbone_name, pretrained_base=True,
                                    norm_layer=self.backbone_norm_layer, **self._kwargs)

        backbone_state_dict = self.backbone.state_dict()
        pretrained_state_dict = pretrained.state_dict()

        backbone_state_dict.update(pretrained_state_dict)
        self.backbone.load_state_dict(backbone_state_dict)

        if self.inference_mode:
            for param in self.backbone.parameters():
                param.requires_grad = False

    def set_prediction_mode(self):
        self.inference_mode = True
        self.eval()

    def forward(self, x, additional_features=None):
        with ExitStack() as stack:
            if self.inference_mode:
                stack.enter_context(paddle.no_grad())

            c1, _, c3, c4 = self.backbone(x, additional_features)
            c1 = self.skip_project(c1)

            x = self.aspp(c4)
            x = F.interpolate(x, c1.shape[2:], mode='bilinear', align_corners=True)
            x = paddle.concat((x, c1), axis=1)
            x = self.head(x)

        return x,


class _SkipProject(nn.Layer):
    def __init__(self, in_channels, out_channels, norm_layer=nn.BatchNorm2D):
        super(_SkipProject, self).__init__()

        self.skip_project = nn.Sequential(
            nn.Conv2D(in_channels, out_channels, kernel_size=1, bias_attr=False),
            norm_layer(out_channels),
            nn.ReLU()
        )

    def forward(self, x):
        return self.skip_project(x)


class _DeepLabHead(nn.Layer):
    def __init__(self, out_channels, in_channels, mid_channels=256, norm_layer=nn.BatchNorm2D):
        super(_DeepLabHead, self).__init__()

        self.block = nn.Sequential(
            SeparableConv2d(in_channels=in_channels, out_channels=mid_channels, dw_kernel=3,
                            dw_padding=1, norm_layer=norm_layer),
            SeparableConv2d(in_channels=mid_channels, out_channels=mid_channels, dw_kernel=3,
                            dw_padding=1, norm_layer=norm_layer),
            nn.Conv2D(in_channels=mid_channels, out_channels=out_channels, kernel_size=1)
        )

    def forward(self, x):
        return self.block(x)


class _ASPP(nn.Layer):
    def __init__(self, in_channels, atrous_rates, out_channels=256,
                 project_dropout=0.5, norm_layer=nn.BatchNorm2D):
        super(_ASPP, self).__init__()

        b0 = nn.Sequential(
            nn.Conv2D(in_channels=in_channels, out_channels=out_channels, kernel_size=1, bias_attr=False),
            norm_layer(out_channels),
            nn.ReLU()
        )

        rate1, rate2, rate3 = tuple(atrous_rates)
        b1 = _ASPPConv(in_channels, out_channels, rate1, norm_layer)
        b2 = _ASPPConv(in_channels, out_channels, rate2, norm_layer)
        b3 = _ASPPConv(in_channels, out_channels, rate3, norm_layer)
        b4 = _AsppPooling(in_channels, out_channels, norm_layer=norm_layer)

        self.concurent = nn.LayerList([b0, b1, b2, b3, b4])

        project = [
            nn.Conv2D(in_channels=5 * out_channels, out_channels=out_channels,
                      kernel_size=1, bias_attr=False),
            norm_layer(out_channels),
            nn.ReLU()
        ]
        if project_dropout > 0:
            project.append(nn.Dropout(project_dropout))
        self.project = nn.Sequential(*project)

    def forward(self, x):
        x = paddle.concat([block(x) for block in self.concurent], axis=1)

        return self.project(x)


class _AsppPooling(nn.Layer):
    def __init__(self, in_channels, out_channels, norm_layer):
        super(_AsppPooling, self).__init__()

        self.gap = nn.Sequential(
            nn.AdaptiveAvgPool2D((1, 1)),
            nn.Conv2D(in_channels=in_channels, out_channels=out_channels,
                      kernel_size=1, bias_attr=False),
            norm_layer(out_channels),
            nn.ReLU()
        )

    def forward(self, x):
        pool = self.gap(x)
        return F.interpolate(pool, x.shape[2:], mode='bilinear', align_corners=True)


def _ASPPConv(in_channels, out_channels, atrous_rate, norm_layer):
    block = nn.Sequential(
        nn.Conv2D(in_channels=in_channels, out_channels=out_channels,
                  kernel_size=3, padding=atrous_rate,
                  dilation=atrous_rate, bias_attr=False),
        norm_layer(out_channels),
        nn.ReLU()
    )

    return block



import paddle
import paddle.nn.functional as F
import numpy as np
from scipy.optimize import fmin_l_bfgs_b

from .base import BasePredictor


class BRSBasePredictor(BasePredictor):
    def __init__(self, model, opt_functor, optimize_after_n_clicks=1, **kwargs):
        super().__init__(model, **kwargs)
        self.optimize_after_n_clicks = optimize_after_n_clicks
        self.opt_functor = opt_functor

        self.opt_data = None
        self.input_data = None

    def set_input_image(self, image_nd):
        super().set_input_image(image_nd)
        self.opt_data = None
        self.input_data = None

    def _get_clicks_maps_nd(self, clicks_lists, image_shape, radius=1):
        pos_clicks_map = np.zeros([len(clicks_lists), 1] + image_shape, dtype=np.float32)
        neg_clicks_map = np.zeros([len(clicks_lists), 1] + image_shape, dtype=np.float32)

        for list_indx, clicks_list in enumerate(clicks_lists):
            for click in clicks_list:
                y, x = click.coords
                y, x = int(round(y)), int(round(x))
                y1, x1 = y - radius, x - radius
                y2, x2 = y + radius + 1, x + radius + 1

                if click.is_positive:
                    pos_clicks_map[list_indx, 0, y1:y2, x1:x2] = True
                else:
                    neg_clicks_map[list_indx, 0, y1:y2, x1:x2] = True

        with paddle.no_grad():
            pos_clicks_map = paddle.to_tensor(pos_clicks_map)
            neg_clicks_map = paddle.to_tensor(neg_clicks_map)
        return pos_clicks_map, neg_clicks_map

    def get_states(self):
        return {'transform_states': self._get_transform_states(), 'opt_data': self.opt_data}

    def set_states(self, states):
        self._set_transform_states(states['transform_states'])
        self.opt_data = states['opt_data']


class FeatureBRSPredictor(BRSBasePredictor):
    def __init__(self, model, opt_functor, insertion_mode='after_deeplab', **kwargs):
        super().__init__(model, opt_functor=opt_functor, **kwargs)
        self.insertion_mode = insertion_mode
        self._c1_features = None
        self.model = model

        if self.insertion_mode == 'after_deeplab':
            self.num_channels = model.feature_extractor.ch
        elif self.insertion_mode == 'after_c4':
            self.num_channels = model.feature_extractor.aspp_in_channels
        elif self.insertion_mode == 'after_aspp':
            self.num_channels = model.feature_extractor.ch + 32
        else:
            raise NotImplementedError

    def _get_prediction(self, image_nd, clicks_lists, is_image_changed):
        points_nd = self.get_points_nd(clicks_lists)
        pos_mask, neg_mask = self._get_clicks_maps_nd(clicks_lists, image_nd.shape[2:])

        num_clicks = len(clicks_lists[0])
        bs = image_nd.shape[0] // 2 if self.with_flip else image_nd.shape[0]

        if self.opt_data is None or self.opt_data.shape[0] // (2 * self.num_channels) != bs:
            self.opt_data = np.zeros((bs * 2 * self.num_channels), dtype=np.float64)

        if num_clicks <= self.net_clicks_limit or is_image_changed or self.input_data is None:
            self.input_data = self._get_head_input(image_nd, points_nd)

        def get_prediction_logits(scale, bias):
            scale = scale.reshape([bs, -1, 1, 1])
            bias = bias.reshape([bs, -1, 1, 1])
            if self.with_flip:
                scale = scale.tile([2, 1, 1, 1])
                bias = bias.tile([2, 1, 1, 1])

            scaled_backbone_features = self.input_data * scale
            scaled_backbone_features = scaled_backbone_features + bias
            if self.insertion_mode == 'after_c4':
                x = self.net.feature_extractor.aspp(scaled_backbone_features)
                x = F.interpolate(x, mode='bilinear', size=self._c1_features.shape[2:], align_corners=True)
                x = paddle.concat((x, self._c1_features), axis=1)
                scaled_backbone_features = self.net.feature_extractor.head(x)

            elif self.insertion_mode == 'after_aspp':
                scaled_backbone_features = self.net.feature_extractor.head(scaled_backbone_features)

            pred_logits = self.net.head(scaled_backbone_features)
            pred_logits = F.interpolate(pred_logits, size=image_nd.shape[2:], mode='bilinear',
                                        align_corners=True)
            return pred_logits

        self.opt_functor.init_click(get_prediction_logits, pos_mask, neg_mask)
        if num_clicks > self.optimize_after_n_clicks:
            opt_result = fmin_l_bfgs_b(func=self.opt_functor, x0=self.opt_data,
                                       **self.opt_functor.optimizer_params)
            self.opt_data = opt_result[0]
        with paddle.no_grad():
            if self.opt_functor.best_prediction is not None:
                opt_pred_logits = self.opt_functor.best_prediction
            else:
                opt_data_nd = paddle.to_tensor(self.opt_data)
                scale, bias, _ = self.opt_functor.unpack_opt_params(opt_data_nd)
                opt_pred_logits = get_prediction_logits(scale, bias)

        return opt_pred_logits

    def _get_head_input(self, image_nd, points):
        with paddle.no_grad():
            image_nd, prev_mask = self.net.prepare_input(image_nd)
            coord_features = self.net.get_coord_features(image_nd, prev_mask, points)

            if self.net.rgb_conv is not None:
                x = self.net.rgb_conv(paddle.concat((image_nd, coord_features), axis=1))
                additional_features = None
            elif hasattr(self.net, 'maps_transform'):
                x = image_nd
                additional_features = self.net.maps_transform(coord_features)

            if self.insertion_mode == 'after_c4' or self.insertion_mode == 'after_aspp':
                c1, _, c3, c4 = self.net.feature_extractor.backbone(x, additional_features)
                c1 = self.net.feature_extractor.skip_project(c1)

                if self.insertion_mode == 'after_aspp':
                    x = self.net.feature_extractor.aspp(c4)
                    x = F.interpolate(x, size=c1.shape[2:], mode='bilinear', align_corners=True)
                    x = paddle.concat((x, c1), axis=1)
                    backbone_features = x
                else:
                    backbone_features = c4
                    self._c1_features = c1
            else:
                backbone_features = self.net.feature_extractor(x, additional_features)[0]

            return backbone_features


class HRNetFeatureBRSPredictor(BRSBasePredictor):
    def __init__(self, model, opt_functor, insertion_mode='A', **kwargs):
        super().__init__(model, opt_functor=opt_functor, **kwargs)
        self.insertion_mode = insertion_mode
        self._c1_features = None
        # sself.model = model

        if self.insertion_mode == 'A':
            self.num_channels = sum(k * model.feature_extractor.width for k in [1, 2, 4, 8])
        elif self.insertion_mode == 'C':
            self.num_channels = 2 * model.feature_extractor.ocr_width
        else:
            raise NotImplementedError

    def _get_prediction(self, image_nd, clicks_lists, is_image_changed):
        points_nd = self.get_points_nd(clicks_lists)
        pos_mask, neg_mask = self._get_clicks_maps_nd(clicks_lists, image_nd.shape[2:])
        num_clicks = len(clicks_lists[0])
        bs = image_nd.shape[0] // 2 if self.with_flip else image_nd.shape[0]

        if self.opt_data is None or self.opt_data.shape[0] // (2 * self.num_channels) != bs:
            self.opt_data = np.zeros((bs * 2 * self.num_channels), dtype=np.float64)

        if num_clicks <= self.net_clicks_limit or is_image_changed or self.input_data is None:
            self.input_data = self._get_head_input(image_nd, points_nd)

        def get_prediction_logits(scale, bias):
            scale = scale.reshape([bs, -1, 1, 1])
            bias = bias.reshape([bs, -1, 1, 1])
            if self.with_flip:
                scale = scale.tile([2, 1, 1, 1])
                bias = bias.tile([2, 1, 1, 1])

            scaled_backbone_features = self.input_data * scale
            scaled_backbone_features = scaled_backbone_features + bias

            if self.insertion_mode == 'A':
                if self.net.feature_extractor.ocr_width > 0:
                    out_aux = self.net.feature_extractor.aux_head(scaled_backbone_features)
                    out_aux.stop_gradient = False
                    feats = self.net.feature_extractor.conv3x3_ocr(scaled_backbone_features)
                    feats.stop_gradient = False
                    context = self.net.feature_extractor.ocr_gather_head(feats, out_aux)
                    context.stop_gradient = False
                    feats = self.net.feature_extractor.ocr_distri_head(feats, context)
                    feats.stop_gradient = False
                else:
                    feats = scaled_backbone_features
                pred_logits = self.net.feature_extractor.cls_head(feats)
                pred_logits.stop_gradient = False
            elif self.insertion_mode == 'C':
                pred_logits = self.net.feature_extractor.cls_head(scaled_backbone_features)
            else:
                raise NotImplementedError

            pred_logits = F.interpolate(pred_logits, size=image_nd.shape[2:], mode='bilinear',
                                        align_corners=True)
            return pred_logits

        self.opt_functor.init_click(get_prediction_logits, pos_mask, neg_mask)
        if num_clicks > self.optimize_after_n_clicks:
            opt_result = fmin_l_bfgs_b(func=self.opt_functor, x0=self.opt_data,
                                       **self.opt_functor.optimizer_params)
            self.opt_data = opt_result[0]

        with paddle.no_grad():
            if self.opt_functor.best_prediction is not None:
                opt_pred_logits = self.opt_functor.best_prediction
            else:
                opt_data_nd = paddle.to_tensor(self.opt_data)
                opt_vars, _ = self.opt_functor.unpack_opt_params(opt_data_nd)
                opt_pred_logits = get_prediction_logits(*opt_vars)

        return opt_pred_logits

    def _get_head_input(self, image_nd, points):
        with paddle.no_grad():
            image_nd, prev_mask = self.net.prepare_input(image_nd)
            coord_features = self.net.get_coord_features(image_nd, prev_mask, points)

            if self.net.rgb_conv is not None:
                x = self.net.rgb_conv(paddle.concat((image_nd, coord_features), axis=1))
                additional_features = None
            else:
                x = image_nd
                additional_features = self.net.maps_transform(coord_features)

            feats = self.net.feature_extractor.compute_hrnet_feats(x, additional_features)

            if self.insertion_mode == 'A':
                backbone_features = feats
            elif self.insertion_mode == 'C':
                out_aux = self.net.feature_extractor.aux_head(feats)
                feats = self.net.feature_extractor.conv3x3_ocr(feats)

                context = self.net.feature_extractor.ocr_gather_head(feats, out_aux)
                backbone_features = self.net.feature_extractor.ocr_distri_head(feats, context)
            else:
                raise NotImplementedError

        return backbone_features


class InputBRSPredictor(BRSBasePredictor):
    def __init__(self, model, opt_functor, optimize_target='rgb', **kwargs):
        super().__init__(model, opt_functor=opt_functor, **kwargs)
        self.optimize_target = optimize_target

    def _get_prediction(self, image_nd, clicks_lists, is_image_changed):
        points_nd = self.get_points_nd(clicks_lists)
        pos_mask, neg_mask = self._get_clicks_maps_nd(clicks_lists, image_nd.shape[2:])
        num_clicks = len(clicks_lists[0])

        if self.opt_data is None or is_image_changed:
            if self.optimize_target == 'dmaps':
                opt_channels = self.net.coord_feature_ch - 1 if self.net.with_prev_mask else self.net.coord_feature_ch
            else:
                opt_channels = 3

            bs = image_nd.shape[0] // 2 if self.with_flip else image_nd.shape[0]
            self.opt_data = paddle.zeros((bs, opt_channels, image_nd.shape[2], image_nd.shape[3]), dtype='float64')

        def get_prediction_logits(opt_bias):
            input_image, prev_mask = self.net.prepare_input(image_nd)
            dmaps = self.net.get_coord_features(input_image, prev_mask, points_nd)

            if self.optimize_target == 'rgb':
                input_image = input_image + opt_bias

            elif self.optimize_target == 'dmaps':
                if self.net.with_prev_mask:
                    dmaps[:, 1:, :, :] = dmaps[:, 1:, :, :] + opt_bias
                else:
                    dmaps = dmaps + opt_bias

            if self.net.rgb_conv is not None:
                x = self.net.rgb_conv(paddle.concat((input_image, dmaps), axis=1))
                if self.optimize_target == 'all':
                    x = x + opt_bias
                coord_features = None
            elif hasattr(self.net, 'maps_transform'):
                x = input_image
                coord_features = self.net.maps_transform(dmaps)

            pred_logits = self.net.backbone_forward(x, coord_features=coord_features)['instances']
            pred_logits = F.interpolate(pred_logits, size=image_nd.shape[2:], mode='bilinear', align_corners=True)

            return pred_logits

        self.opt_functor.init_click(get_prediction_logits, pos_mask, neg_mask, shape=self.opt_data.shape)
        if num_clicks > self.optimize_after_n_clicks:
            opt_result = fmin_l_bfgs_b(func=self.opt_functor, x0=self.opt_data.cpu().numpy().ravel(),
                                       **self.opt_functor.optimizer_params)

            self.opt_data = paddle.to_tensor(opt_result[0])
            self.opt_data = paddle.reshape(self.opt_data, shape=self.opt_data.shape)

        with paddle.no_grad():
            if self.opt_functor.best_prediction is not None:
                opt_pred_logits = self.opt_functor.best_prediction
            else:
                opt_vars, _ = self.opt_functor.unpack_opt_params(self.opt_data)
                opt_pred_logits = get_prediction_logits(*opt_vars)

        return opt_pred_logits



import os.path as osp

import paddle

from model.is_hrnet_model import HRNetModel
# from model.is_deeplab_model import DeeplabModel
# from util import model_path

here = osp.dirname(osp.abspath(__file__))


class HRNet18_OCR48:
    name = "HRNet18_OCR48"

    def load_params(self, params_path):
        model = HRNetModel(
            width=18, 
            ocr_width=48, 
            small=True, 
            with_aux_output=True, 
            use_rgb_conv=False,
            use_leaky_relu=True, 
            use_disks=True, 
            with_prev_mask=True, 
            norm_radius=5, 
            cpu_dist_maps=False
        )
        para_state_dict = paddle.load(params_path)
        model.set_dict(para_state_dict)
        model.eval()
        return model


class HRNet18_OCR64:
    name = "HRNet18_OCR64"

    def load_params(self, params_path):
        model = HRNetModel(
            width=18, 
            ocr_width=64, 
            small=False, 
            with_aux_output=True, 
            use_leaky_relu=True,
            use_rgb_conv=False,
            use_disks=True, 
            norm_radius=5, 
            with_prev_mask=True,
            cpu_dist_maps=False  # 目前打包cython有些问题，先默认用False
        )
        para_state_dict = paddle.load(params_path)
        model.set_dict(para_state_dict)
        model.eval()
        return model


models = [HRNet18_OCR48(), HRNet18_OCR64()]


def findModelbyName(model_name):
    for idx, mt in enumerate(models):
        if model_name == mt.name:
            return models[idx], idx


import paddle.nn as nn


class ConvHead(nn.Layer):
    def __init__(self, out_channels, in_channels=32, num_layers=1,
                 kernel_size=3, padding=1,
                 norm_layer=nn.BatchNorm2D):
        super(ConvHead, self).__init__()
        convhead = []

        for i in range(num_layers):
            convhead.extend([
                nn.Conv2D(in_channels, in_channels, kernel_size, padding=padding),
                nn.ReLU(),
                norm_layer(in_channels)
            ])
        convhead.append(nn.Conv2D(in_channels, out_channels, 1, padding=0))

        self.convhead = nn.Sequential(*convhead)

    def forward(self, *inputs):
        return self.convhead(inputs[0])


class SepConvHead(nn.Layer):
    def __init__(self, num_outputs, in_channels, mid_channels, num_layers=1,
                 kernel_size=3, padding=1, dropout_ratio=0.0, dropout_indx=0,
                 norm_layer=nn.BatchNorm2D):
        super(SepConvHead, self).__init__()

        sepconvhead = []

        for i in range(num_layers):
            sepconvhead.append(
                SeparableConv2d(in_channels=in_channels if i == 0 else mid_channels,
                                out_channels=mid_channels,
                                dw_kernel=kernel_size, dw_padding=padding,
                                norm_layer=norm_layer)
            )
            if dropout_ratio > 0 and dropout_indx == i:
                sepconvhead.append(nn.Dropout(dropout_ratio))

        sepconvhead.append(
            nn.Conv2D(in_channels=mid_channels, out_channels=num_outputs, kernel_size=1, padding=0)
        )

        self.layers = nn.Sequential(*sepconvhead)

    def forward(self, *inputs):
        x = inputs[0]

        return self.layers(x)


class SeparableConv2d(nn.Layer):
    def __init__(self, in_channels, out_channels, dw_kernel, dw_padding, dw_stride=1, use_bias=False, norm_layer=None):
        super(SeparableConv2d, self).__init__()
        self.body = nn.Sequential(
            nn.Conv2D(in_channels, in_channels, kernel_size=dw_kernel, stride=dw_stride,
                      padding=dw_padding, bias_attr=use_bias, groups=in_channels),
            nn.Conv2D(in_channels, out_channels, kernel_size=1, stride=1, bias_attr=use_bias),
            norm_layer(out_channels),
            nn.ReLU()
        )

    def forward(self, x):
        return self.body(x)


import os.path as osp
import random


class ColorMask(object):
    def __init__(
        self, color_path, shuffle=False
    ):
        self.color_maps = []
        self.index = 0
        with open(color_path, "r") as f:
            self.color_maps = f.readlines()
        if shuffle:
            random.shuffle(self.color_maps)
        self.color_map_nums = len(self.color_maps)

    def get_color(self):
        color = self.color_maps[self.index].strip()
        self.index += 1
        if self.index == self.color_map_nums:
            self.index = 0
        return self.to_list(color)

    def to_list(self, color):
        r, g, b = color.split(",")
        return [int(r), int(g), int(b)]



import paddle
import paddle.nn as nn
import numpy as np
import model.initializer as initializer


class DistMaps(nn.Layer):
    def __init__(self, norm_radius, spatial_scale=1.0, cpu_mode=True, use_disks=False):
        super(DistMaps, self).__init__()
        self.spatial_scale = spatial_scale
        self.norm_radius = norm_radius
        self.cpu_mode = cpu_mode
        self.use_disks = use_disks

        if self.cpu_mode:
            from util.cython import get_dist_maps
            self._get_dist_maps = get_dist_maps

    def get_coord_features(self, points, batchsize, rows, cols):
        if self.cpu_mode:
            coords = []
            for i in range(batchsize):
                norm_delimeter = 1.0 if self.use_disks else self.spatial_scale * self.norm_radius
                coords.append(self._get_dist_maps(points[i].numpy().astype('float32'), rows, cols,
                                                  norm_delimeter))
            coords = paddle.to_tensor(np.stack(coords, axis=0)).astype('float32')
        else:
            num_points = points.shape[1] // 2
            points = points.reshape([-1, points.shape[2]])
            points, points_order = paddle.split(points, [2, 1], axis=1)
            invalid_points = paddle.max(points, axis=1, keepdim=False) < 0
            row_array = paddle.arange(start=0, end=rows, step=1, dtype='float32')
            col_array = paddle.arange(start=0, end=cols, step=1, dtype='float32')

            coord_rows, coord_cols = paddle.meshgrid(row_array, col_array)
            coords = paddle.unsqueeze(paddle.stack([coord_rows, coord_cols], axis=0), axis=0).tile(
                [points.shape[0], 1, 1, 1])
            
            add_xy = (points * self.spatial_scale).reshape([points.shape[0], points.shape[1], 1, 1])
            coords = coords - add_xy
            if not self.use_disks:
                coords = coords / (self.norm_radius * self.spatial_scale)

            coords = coords * coords
            coords[:, 0] += coords[:, 1]
            coords = coords[:, :1]
            invalid_points = invalid_points.numpy()

            coords[invalid_points, :, :, :] = 1e6
            coords = coords.reshape([-1, num_points, 1, rows, cols])
            coords = paddle.min(coords, axis=1)
            coords = coords.reshape([-1, 2, rows, cols])

        if self.use_disks:
            coords = (coords <= (self.norm_radius * self.spatial_scale) ** 2).astype('float32')
        else:
            coords = paddle.tanh(paddle.sqrt(coords) * 2)
        return coords

    def forward(self, x, coords):
        return self.get_coord_features(coords, x.shape[0], x.shape[2], x.shape[3])


class ScaleLayer(nn.Layer):
    def __init__(self, init_value=1.0, lr_mult=1):
        super().__init__()
        self.lr_mult = lr_mult
        self.scale = self.create_parameter(shape=[1],
                                           dtype='float32',
                                           default_initializer=nn.initializer.Constant(init_value / lr_mult))

    def forward(self, x):
        scale = paddle.abs(self.scale * self.lr_mult)
        return x * scale


class BatchImageNormalize:
    def __init__(self, mean, std):
        self.mean = paddle.to_tensor(np.array(mean)[np.newaxis, :, np.newaxis, np.newaxis]).astype('float32')
        self.std = paddle.to_tensor(np.array(std)[np.newaxis, :,np.newaxis, np.newaxis]).astype('float32')

    def __call__(self, tensor):
        tensor = (tensor - self.mean) / self.std
        return tensor
